#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include "SDL2_gfxPrimitives.h"

#include <algorithm>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iomanip>
#include <list>
#include <sstream>
#include <string>
#include <vector>

#include "text.h"
#include "SoundPlayer.h"
#include "MusicPlayer.h"

/* Global SDL variables */
SDL_Event ev;
SDL_Renderer *sdlRenderer;
SDL_TimerID sdlTimer;

/* Global variables for sound playback */
SoundPlayer sound_success("242501__gabrielaraujo__powerup-success.wav");
SoundPlayer sound_error("142608__autistic-lucario__error_-12.wav");
SoundPlayer sound_magic("22267__zeuss__the-chime.wav");
MusicPlayer music("A Himitsu - Adventures [Argofox].mp3");

/* Dimensions of the board */
const int COLS = 6; // (4-30)
const int ROWS = 8; // (8-20)

/* Some other size parameters */
const int SQUARE = 50; // Size of a square that contains a gem
const int MARGIN = 40; // Distance between edge of window and board

/* Dimensions of game window
 * Determined by the settings above */
const int WIDTH = COLS*SQUARE + MARGIN*2 + 320;
const int HEIGHT = ROWS*SQUARE + MARGIN*2;


/**
 * @class Background
 * @author zsmb
 * @brief A class that can load an image from file and then paint it on
 * the screen using an SDL_Renderer. It can also draw just a given width
 * of the background, starting from the left edge of the screen.
 */
class Background {
    SDL_Texture* bg_texture; // For storing the loaded texture
    
public:
    /**
     * @brief Background ctor, filename is hardcoded for now
     */
    Background() {
        bg_texture = IMG_LoadTexture(sdlRenderer, "bg.jpg");
        if(bg_texture == nullptr) {
            std::cerr << "couldn't load bg texture!!!" << std::endl;
        }
    }
    
    /**
     * @brief Background dtor, frees up resources
     */
    ~Background() {
        SDL_DestroyTexture(bg_texture);
    }
    
    /**
     * @brief Draws the stored background, stretched to the dimensions
     * of the screen
     */
    void draw() {
        boxColor(sdlRenderer, 0, 0, WIDTH, HEIGHT, 0xAAAAAAFF);
        SDL_Rect dest = {0, 0, WIDTH, HEIGHT};
        SDL_RenderCopy(sdlRenderer, bg_texture, nullptr, &dest);
        boxColor(sdlRenderer, 0, 0, WIDTH, HEIGHT, 0x00000066);
    }
    
    /**
     * @brief Draws the background left to right until a certain x value
     * @param x_bound the limit to draw until
     */
    void draw_in_part(int x_bound) {
        boxColor(sdlRenderer, 0, 0, x_bound-1, HEIGHT, 0xAAAAAAFF);
        double ratio = (double)x_bound/WIDTH;
        
        int texture_width;
        int texture_height;
        SDL_QueryTexture(bg_texture, nullptr, nullptr, &texture_width, &texture_height);
        
        SDL_Rect src = {0, 0, int(ratio*texture_width), texture_height};
        SDL_Rect dest = {0, 0, x_bound, HEIGHT};
        SDL_RenderCopy(sdlRenderer, bg_texture, &src, &dest);
        boxColor(sdlRenderer, 0, 0, x_bound-1, HEIGHT, 0x00000066);
    }
};


/**
 * @class Gem
 * @author zsmb
 * @brief Abstract base class for the ingame gems
 */
class Gem {
    bool used = false; // Flag for marking that the gem has been already a
                       // part of a match
    
public:    
    /**
     * @brief Marks a gem as have been part of a match
     */
    void set_used() {
        used = true;
    }
    
    /**
     * @brief Gets the status of a gem
     * @return true, if it has been matched
     */
    bool is_used() {
        return used;
    }
    
    /**
     * @brief Draws the gem on the screen
     * @param xpos horizontal center of the gem
     * @param ypos vertical center of the gem
     */
    virtual void draw(int xpos, int ypos) = 0;
    
    /**
     * @brief Returns the type of gem (matching gems should have the same type)
     * @return the gem's type
     */
    virtual int get_type() = 0;
    
    /**
     * @brief Virtual dtor
     */
    virtual ~Gem() {};
};


/**
 * @class RedGem
 * @author zsmb
 * @brief A red, sort of bell-looking gem
 */
class RedGem : public Gem {
    int colour = 0xEE1E2AFF;
    
    void draw(int xpos, int ypos) {
        boxColor(sdlRenderer, xpos+4, ypos+4, xpos+SQUARE-5, ypos+SQUARE-5, 0x00000044);
        
        Sint16 x = Sint16(xpos + SQUARE/2);
        Sint16 y = Sint16(ypos + SQUARE/2);
        Sint16 sqp3 = Sint16(SQUARE/3);
        Sint16 sqp4 = Sint16(SQUARE/4);
        Sint16 sqp6 = Sint16(SQUARE/6);
        Sint16 sqp12 = Sint16(SQUARE/8);
        
        Sint16 vx[6] = {Sint16(x-sqp4), Sint16(x-sqp12), x, Sint16(x+sqp12), Sint16(x+sqp4), x};
        Sint16 vy[6] = {Sint16(y+sqp6), Sint16(y-sqp4), Sint16(y-sqp3), Sint16(y-sqp4), Sint16(y+sqp6), Sint16(y+sqp3)};
        
        filledPolygonColor(sdlRenderer, vx, vy, 6, colour);
        aapolygonColor(sdlRenderer, vx, vy, 6, colour);
    }
    
    int get_type() {
        return 1;
    }
};

/**
 * @class BlueGem
 * @author zsmb
 * @brief A blue, circular gem
 */
class BlueGem : public Gem {
    int colour = 0x2255FFFF;
    
    void draw(int xpos, int ypos) {
        boxColor(sdlRenderer, xpos+4, ypos+4, xpos+SQUARE-5, ypos+SQUARE-5, 0x00000044);
        filledCircleColor(sdlRenderer, xpos+SQUARE/2, ypos+SQUARE/2, SQUARE/3, colour);
        aacircleColor(sdlRenderer, xpos+SQUARE/2, ypos+SQUARE/2, SQUARE/3, colour);
    }
    
    int get_type() {
        return 2;
    }
};

/**
 * @class GreenGem
 * @author zsmb
 * @brief A green, diamond shaped gem
 */
class GreenGem : public Gem {
    int colour = 0x00CC29FF;
    
    void draw(int xpos, int ypos) {
        boxColor(sdlRenderer, xpos+4, ypos+4, xpos+SQUARE-5, ypos+SQUARE-5, 0x00000044);
        
        Sint16 x = Sint16(xpos + SQUARE/2);
        Sint16 y = Sint16(ypos + SQUARE/2);
        Sint16 sqp3 = Sint16(SQUARE/3);
        
        Sint16 vx[4] = {x, Sint16(x+sqp3), x, Sint16(x-sqp3)};
        Sint16 vy[4] = {Sint16(y-sqp3), y, Sint16(y+sqp3), y};
        
        filledPolygonColor(sdlRenderer, vx, vy, 4, colour);
    }
    
    int get_type() {
        return 3;
    }
};

/**
 * @class PurpleGem
 * @author zsmb
 * @brief A purple, square shaped gem
 */
class PurpleGem : public Gem {
    int colour = 0xCB4EFBFF;
        
    void draw(int xpos, int ypos) {
        boxColor(sdlRenderer, xpos+4, ypos+4, xpos+SQUARE-5, ypos+SQUARE-5, 0x00000044);
        
        int x = xpos+SQUARE/2;
        int y = ypos+SQUARE/2;
        int sq2p7 = SQUARE*2/7;
        
        boxColor(sdlRenderer, x-sq2p7, y-sq2p7, x+sq2p7, y+sq2p7, colour);
    }
    
    int get_type() {
        return 4;
    }
};

/**
 * @class YellowGem
 * @author zsmb
 * @brief A yellow, pentagonal gem
 */
class YellowGem : public Gem {
    int colour = 0xFFC732FF;
    
    void draw(int xpos, int ypos) {        
        boxColor(sdlRenderer, xpos+4, ypos+4, xpos+SQUARE-5, ypos+SQUARE-5, 0x00000044);
        
        Sint16 x = Sint16(xpos + SQUARE/2);
        Sint16 y = Sint16(ypos + SQUARE*.55);
        Sint16 sqp3 = Sint16(SQUARE/3);
        Sint16 sqp4 = Sint16(SQUARE/4);
        Sint16 sq2p11 = Sint16(SQUARE*2/11);
        Sint16 sqp10 = Sint16(SQUARE/10);
        Sint16 sq7p20 = Sint16(SQUARE*7/20);
        
        
        Sint16 vx[5] = {x, Sint16(x+sqp3), Sint16(x+sq2p11), Sint16(x-sq2p11), Sint16(x-sqp3)};
        Sint16 vy[5] = {Sint16(y-sq7p20), Sint16(y-sqp10), Sint16(y+sqp4), Sint16(y+sqp4), Sint16(y-sqp10)};
        Sint16 vy_aa1[5] = {Sint16(y-sq7p20), Sint16(y-sqp10), Sint16(y+sqp4-1), Sint16(y+sqp4), Sint16(y-sqp10)};
        Sint16 vy_aa2[5] = {Sint16(y-sq7p20), Sint16(y-sqp10), Sint16(y+sqp4), Sint16(y+sqp4-1), Sint16(y-sqp10)};
        
        filledPolygonColor(sdlRenderer, vx, vy, 5, colour);
        aapolygonColor(sdlRenderer, vx, vy_aa1, 5, colour);
        aapolygonColor(sdlRenderer, vx, vy_aa2, 5, colour);
    }
    
    int get_type() {
        return 5;
    }
};

/**
 * @class CyanGem
 * @author zsmb
 * @brief A light blue gemmy gem
 */
class CyanGem : public Gem {
    int colour = 0x00E0FDFF;
    
    void draw(int xpos, int ypos) {
        boxColor(sdlRenderer, xpos+4, ypos+4, xpos+SQUARE-5, ypos+SQUARE-5, 0x00000044);
        
        Sint16 x = Sint16(xpos + SQUARE/2);
        Sint16 y = Sint16(ypos + SQUARE/2);
        Sint16 sq2p7 = Sint16(SQUARE*2/7);
        Sint16 sq3p11 = Sint16(SQUARE*3/11);
        
        Sint16 vx[6] = {Sint16(x-sq2p7), Sint16(x-sq2p7), Sint16(x+sq2p7-sq3p11), Sint16(x+sq2p7), Sint16(x+sq2p7), Sint16(x-sq2p7+sq3p11)};
        Sint16 vy[6] = {Sint16(y+sq2p7), Sint16(y+sq2p7-sq3p11), Sint16(y-sq2p7), Sint16(y-sq2p7), Sint16(y-sq2p7+sq3p11), Sint16(y+sq2p7)};
        
        filledPolygonColor(sdlRenderer, vx, vy, 6, colour);
    }
    
    int get_type() {
        return 6;
    }
};


/**
 * @class StatCollector
 * @author zsmb
 * @brief A stat for collecting stats about the ingame events,
 * with getters to access them for display
 */
class StatCollector {
    std::vector<int> scores;    // for storing high scores
    int last_score = -1;        // to remember what the last logged score was
    
    std::list<std::pair<int,std::string>> logs; // log messages of matches to display
                                                // entires are of format (age, message)
    
    int current = 0;    // currently accumulated score
    
    /**
     * @brief Gets the scores vector into a consistent state
     */
    void fill_sort_cut() {
        while(scores.size() < 5) {
            scores.push_back(0);
        }
        std::sort(scores.begin(), scores.end(), std::greater<int>());
        while(scores.size() > 5) {
            scores.pop_back();
        }
    }
    
public:
    /**
     * @brief Statcollector ctor, reads scores from file if possible
     */
    StatCollector() {
        std::ifstream stored_stores("scores.dat");
        if(stored_stores.good()) {
            int temp;
            while(stored_stores >> temp) {
                scores.push_back(temp);
            }
        }
        fill_sort_cut();
    }
    
    /**
     * @brief Statcollector dtor, saves scores to file 
     */
    ~StatCollector() {
        std::ofstream stored_scores("scores.dat");
        for(int s : scores) {
            stored_scores << s << " ";
        }
    }
    
    /**
     * @brief Getter function
     * @return the currently accumulated score
     */
    int get_current_score() {
        return current;
    }
    
    /**
     * @brief Getter function
     * @return the last logged score
     */
    int get_last_score() {
        return last_score;
    }
    
    /**
     * @brief Gets read/write access to the stored logs
     * @return a list of log entries
     */
    std::list<std::pair<int, std::string>>& get_logs() {
        return logs;
    }
    
    /**
     * @brief Gets read only access to stored scores
     * @return a vector of the stored scores
     */
    std::vector<int> const& get_scores() {
        return scores;
    }
 
    /**
     * @brief Commits the currently accumulated score
     * to the high scores
     */
    void commit_score() {
        if(current != 0) {
            last_score = current;        
        }
        scores.push_back(current);
        fill_sort_cut();
        current = 0;
    }
    
    /**
     * @brief Empties logs and resets current score
     */
    void reset() {
        current = 0;
        logs.clear();
    }
    
    /**
     * @brief Reports a match of a certain length for which a score
     * is to be added to the currently accumulated score
     * @param match_length the number of gems matched
     */
    void report(int match_length) {
        current += 5*match_length*match_length;
        
        std::stringstream ss;
        ss << "Matched " << match_length << " gems for " << 5*match_length*match_length << " points!"
           << " (+" << std::setprecision(1) << (match_length-2)/2.0 << "s)";
        logs.push_back(std::pair<int,std::string>(0,ss.str()));
        
        while(logs.size() > 10) {
            logs.pop_front();
        }
    }
};

/**
 * @class Board
 * @author zsmb
 * @brief A board that contains and manages gems
 */
class Board {
    Gem* gems[ROWS][COLS];  // matrix of gems
    Background& bg;         // reference to the object drawing the background
                            // behind the board (to force a redraw of it)
    StatCollector& sc;      // StatCollector to log events to
    
    double time_remaining = 0;  // game time remaining, in seconds
    int tick_count = 0;         // helper variable for time counting
    
    
    /**
     * @brief Remove all currently matched gems from the board
     * @return true, if there were matches removed
     */
    bool clear_matches() {
        
        bool match_found = false;
        
        for(int r = 0; r < ROWS; r++) {
            int matched = 1;
            int current = gems[r][0]->get_type();
            
            for(int c = 1; c < COLS; c++) {
                int type = gems[r][c]->get_type();
                
                if(type == current) {
                    matched++;
                }
                else {
                    if(matched >= 3) {
                        process_match(r, c-1-matched+1, r, c-1);
                        match_found = true;
                    }
                    matched = 1;
                    current = type;
                }
            }
            
            if(matched >= 3) {
                process_match(r, COLS-1-matched+1, r, COLS-1);
                match_found = true;
            }
        }
        
        for(int c = 0; c < COLS; c++) {
            int matched = 1;
            int current = gems[0][c]->get_type();
            
            for(int r = 1; r < ROWS; r++) {
                int type = gems[r][c]->get_type();
                
                if(type == current) {
                    matched++;
                }
                else {
                    if(matched >= 3) {
                        process_match(r-1-matched+1, c, r-1, c);
                        match_found = true;
                    }
                    matched = 1;
                    current = type;
                }
            }
            
            if(matched >= 3) {
                process_match(ROWS-1-matched+1, c, ROWS-1, c);
                match_found = true;
            }
        }
        
        remove_used();
        
        return match_found;
    }    
    
    /**
     * @brief Fills up each column with new, random gems (only from the top)
     */
    void fill() {
        for(int c = 0; c < COLS; c++) {
            int r = 0;
            while(r < ROWS && gems[r][c] == nullptr) {
                gems[r][c] = random_gem();
                r++;
            }
        }
    }
    
    /**
     * @brief Checks if there's a match currently on the board
     * @return true, if there is a match currently on the board
     */
    bool match_present() {
        for(int r = 0; r < ROWS; r++) {
            int matched = 1;
            int current = gems[r][0]->get_type();
            
            for(int c = 1; c < COLS; c++) {
                int type = gems[r][c]->get_type();
                if(type == current) {
                    matched++;
                    if(matched >= 3) {
                        return true;
                    }
                }
                else {
                    current = type;
                    matched = 1;
                }
            }
        }
        
        for(int c = 0; c < COLS; c++) {
            int matched = 1;
            int current = gems[0][c]->get_type();
            
            for(int r = 1; r < ROWS; r++) {
                int type = gems[r][c]->get_type();
                if(type == current) {
                    matched++;
                    if(matched >= 3) {
                        return true;
                    }
                }
                else {
                    current = type;
                    matched = 1;
                }
            }
        }
        
        return false;
    }
    
    /**
     * @brief Puts two integers in increasing order
     * @param a first integer
     * @param b second integer
     */
    void order(int& a, int&b) {
        if(b < a) {
            int temp = a;
            a = b;
            b = temp;
        }
    }
    
     /**
     * @brief Log the occurance of a match, notify matched gems
     * @param r1 row the match starts in
     * @param c1 column the match starts in
     * @param r2 row the match ends in
     * @param c2 column the match ends in
     */
    void process_match(int r1, int c1, int r2, int c2) {       
        order(r1, r2);
        order(c1, c2);
        
        for(int r = r1; r <= r2; r++) {
            for(int c = c1; c <= c2; c++) {
                gems[r][c]->set_used();
            }
        }
        
        int match_length;
        
        if(r1 == r2) {
            match_length = c2-c1+1;
        }
        else if(c1 == c2) {
            match_length = r2-r1+1;
        }
        else {
            std::cerr << "This can not possibly happen" << std::endl;
        }
        
        sc.report(match_length);
        time_remaining += (match_length-2)/2.0;
    }
    
    /**
     * @brief Returns a random gem
     * @return the created gem
     */
    Gem* random_gem() {
        switch(rand()%6) {
            case 0:
                return new YellowGem();
            case 1:
                return new PurpleGem();
            case 2:
                return new GreenGem();
            case 3:
                return new BlueGem();
            case 4:
                return new RedGem();
            case 5:
                return new CyanGem();
        }
        return nullptr;
    }
    
    /**
     * @brief Remove already matched gems from the board
     */
    void remove_used() {
        for(int r = 0; r < ROWS; r++) {
            for(int c = 0; c < COLS; c++) {
                if(gems[r][c]->is_used()) {
                    delete gems[r][c];
                    gems[r][c] = nullptr;
                }
            }
        }
    }
    
    /**
     * @brief Shifts gems in a column down as far as is possible for each gem
     * @param c the column to shift gems in
     */
    void shift_column(int c) {
        int sorted = 0;
        int i = ROWS-1;
        while(i > 0 && sorted < ROWS) {
        
            if(gems[i][c] != nullptr && gems[i][c]->is_used()) {
                delete gems[i][c];
                gems[i][c] = nullptr;
            }
            
            if(gems[i][c] == nullptr) {
                shift_column_onto(c, i);
                sorted++;
            }
            else {
                i--;
            }
        }
    }    
    
    /**
     * @brief Shifts all gems above a given index down by one
     * @param c the column to operate in
     * @param empty_space_index the empty space in the column
     */
    void shift_column_onto(int c, int empty_space_index) {
        int n = empty_space_index;
        for(int i = n; i > 0; i--) {
            gems[i][c] = gems[i-1][c];
        }
        gems[0][c] = nullptr;
    }
    
    /**
     * @brief Drop all gems as far down as possible
     */
    void shift() {
        for(int c = 0; c < COLS; c++) {
            shift_column(c);
        }
    }
    
    /**
     * @brief Swaps to gems on the board
     * @param r1 row of the first gem
     * @param c1 column of the first gem
     * @param r2 row of the second gem
     * @param c2 column of the second gem
     */
    void swap(int r1, int c1, int r2, int c2) {
        Gem* temp = gems[r1][c1];
        gems[r1][c1] = gems[r2][c2];
        gems[r2][c2] = temp;
    }
    
    /**
     * @brief Hangs everything for a given time
     * @param ticks the number of ticks (20ms) to wait
     */
    void wait(int ticks) {
        int count = 0;
        while(SDL_WaitEvent(&ev) && count < ticks) {
            if(ev.type == SDL_USEREVENT) {
                count++;
            }
        }
    }
    
public:
    /**
     * @brief Board ctor
     * @param bg background behind the board
     * @param sc StatCollector to log events to
     */
    Board(Background& bg, StatCollector& sc) : bg(bg), sc(sc) {
        for(int r = 0; r < ROWS; r++) {
            for(int c = 0; c < COLS; c++) {
                gems[r][c] = random_gem();
            }
        }
        
        while(clear_matches()) {
            shift();
            fill();
        }
        sc.reset();
    }
    
    /**
     * @brief Board dtor, frees up resources
     */
    ~Board() {
        for(int r = 0; r < ROWS; r++) {
            for(int c = 0; c < COLS; c++) {
                if(gems[r][c] != nullptr) {
                    delete gems[r][c];
                }
            }
        }
    }
    
    /**
     * @brief Getter function
     * @return game time remaining (seconds)
     */
    double get_time_remaining() {
        return time_remaining;
    }

    /**
     * @brief Attempts to swap to gems (this is successful if it results
     * in a match, in this case all resulting matches also get processed)
     * @param r1 row of the first gem
     * @param c1 column of the first gem
     * @param r2 row of the second gem
     * @param c2 column of the second gem
     * @return true, if the swap was successful
     */
    bool attempt_swap(int r1, int c1, int r2, int c2) {
        swap(r1, c1, r2, c2);
        if(clear_matches()) {
            do {
                shift();
                fill();
                bg.draw_in_part(MARGIN+COLS*SQUARE+5);
                draw();
                SDL_RenderPresent(sdlRenderer);
                sound_success.play();
                wait(25);
            }
            while(clear_matches());
            return true;
        }
        sound_error.play();
        swap(r1, c1, r2, c2);
        return false;
    }
    
    /**
     * @brief Notifies the board of time passing
     * @return true, if there's no game time left on the board
     */
    bool decrease_time() {
        tick_count = (tick_count+1) % 25;
        if(tick_count == 0) {
            time_remaining -= 0.5;
            if(time_remaining <= 0) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * @brief Draws the board on the screen
     */
    void draw() {
        int BOARD_WIDTH = COLS*SQUARE;
        int BOARD_HEIGHT = ROWS*SQUARE;
        
        //// Board background
        boxColor(sdlRenderer, MARGIN-2, MARGIN-2, MARGIN+BOARD_WIDTH+2, MARGIN+BOARD_HEIGHT+2, 0x00000033);
        
        // Gems
        for(int r = 0; r < ROWS; r++) {
            for(int c = 0; c < COLS; c++) {
                Gem* temp = gems[r][c];
                if(temp != nullptr && !temp->is_used()) {
                    temp->draw(MARGIN+c*SQUARE, MARGIN+r*SQUARE);
                }
            }
        }
    }

    /**
     * @brief Check if there's a swap that results in a match
     * @return true, if there's a swap that results in a match
     */
    bool match_possible() {
        if(match_present()) {
            return true;
        }
        
        for(int r = 0; r < ROWS; r++) {
            for(int c = 1; c < COLS; c++) {
                swap(r, c-1, r, c);
                bool possible = match_present();
                swap(r, c-1, r, c);
                if(possible) {
                    return true;
                }
            }
        }
        
        for(int c = 0; c < COLS; c++) {
            for(int r = 1; r < ROWS; r++) {
                swap(r-1, c, r, c);
                bool possible = match_present();
                swap(r-1, c, r, c);
                if(possible) {
                    return true;
                }
            }
        }
        
        return false;
    }

    /**
     * @brief Deletes a number of gems from the board
     * @param n the number of gems to delete (approximately)
     */
    void randomly_delete(int n) {
        for(int i = 0; i < n; i++) {
            int r = rand()%ROWS;
            int c = rand()%COLS;
            
            delete gems[r][c];
            gems[r][c] = nullptr;
        }
        
        do {
            shift();
            fill();
        }
        while(clear_matches());
    }    
    
    /**
     * @brief Resets the board for the start of a new game
     */
    void reset() {
        // New random gems
        for(int r = 0; r < ROWS; r++) {
            for(int c = 0; c < COLS; c++) {
                if(gems[r][c] != nullptr) {
                    delete gems[r][c];
                }
                gems[r][c] = random_gem();
            }
        }
        
        // Get rid of all matches
        while(clear_matches()) {
            shift();
            fill();
        }       
        
        // Reset StatCollector
        sc.reset();
        
        // Reset remaining time
        tick_count = 0;
        time_remaining = 180;
        
        // Display board
        draw();
        SDL_RenderPresent(sdlRenderer);
    }
};

/**
 * @class Selector
 * @author zsmb
 * @brief A selector to move around on top of the board
 */
class Selector {
    Board& b;           // the board to navigate
    int r = 0, c = 0;   // current coordinates of the selector
    
    bool first_selected = false;    // flag for whether there's one position selected
    int r_sel, c_sel;               // coodinates of the selected position
    
    /**
     * @brief Tiny helper function for absolute value of an integer
     * @param x the number to get the absolute value of
     * @return -x if x is negative, otherwise x
     */
    int abs(int x) {
        return x<0?-x:x;
    }
    
public:
    /**
     * @brief Selector ctor
     * @param b the board to navigate
     */
    Selector(Board& b) : b(b) {    
    }
    
    /**
     * @brief Draws the selector on the screen
     */
    void draw() {
        int left = MARGIN + c*SQUARE;
        int top = MARGIN + r*SQUARE;
        
        for(int i = 0; i < 3; i++) {
            rectangleColor(sdlRenderer, left+i, top+i, left+SQUARE-i, top+SQUARE-i, 0xFFFF00FF);
        }
        
        if(first_selected) {
            left = MARGIN + c_sel*SQUARE;
            top = MARGIN + r_sel*SQUARE;
            for(int i = 0; i < 3; i++) {
                rectangleColor(sdlRenderer, left+i, top+i, left+SQUARE-i, top+SQUARE-i, 0xFF0000FF);
            }
        }
    }
    
    /**
     * @brief Moves the selector around the board (loops around at edges)
     * @param direction the direction to move in
     */
    void move(int direction) {
        switch(direction) {
            case SDLK_UP:
            case SDLK_w:
                if(r > 0) {
                    r--;
                }
                else {
                    r = ROWS-1;
                }
                break;
            case SDLK_DOWN:
            case SDLK_s:
                r = (r+1)%ROWS;
                break;
            case SDLK_LEFT:
            case SDLK_a:
                if(c > 0) {
                    c--;
                }
                else {
                    c = COLS-1;
                }
                break;
            case SDLK_RIGHT:
            case SDLK_d:
                c = (c+1)%COLS;
                break;
        }
    }
    
    /**
     * @brief Selects the gem in the current position (notifies board if two have been selected)
     */
    void select() {
        if(first_selected) {
            if((r == r_sel && abs(c-c_sel) == 1) ||
                    (c == c_sel && abs(r-r_sel) == 1)) {
                b.attempt_swap(r, c, r_sel, c_sel);
                first_selected = false;
            }
            else {
                sound_error.play();
                r_sel = 0;
                c_sel = 0;
                first_selected = false;
            }
        }
        else {
            r_sel = r;
            c_sel = c;
            first_selected = true;
        }
    }
    
    /**
     * @brief Deletes current selection
     * @return true if there was a selection to delete
     */
    bool unselect() {
        if(first_selected) {
            first_selected = false;
            return true;
        }
        return false;
    }
    
    
};

/**
 * @class Display
 * @author zsmb
 * @brief Displays various kinds of informations for the player
 */
class Display {
    Board& b;           // the Board the player is using
    StatCollector& sc;  // the Statcollector storing stats
    
    std::vector<std::string> help_lines;    // cached lines of the help file
    
    /* Font settings */
    int small = 14;
    int medium = 20;
    int large = 24;
    int font_colour = 0xFFFFFFFF;
    
    int xpos = MARGIN*2 + COLS*SQUARE; // left edge of the display on screen
    
public:
    /**
     * @brief Display ctor, reads help contents from file (hardcoded for now)
     * @param b active Board
     * @param sc active StatCollector
     */
    Display(Board& b, StatCollector& sc) : b(b), sc(sc) {
        std::ifstream help_file("help.txt");
        if(help_file.good()) {
            std::string temp;
            while(getline(help_file, temp)) {
                help_lines.push_back(temp);
            }
        }
    }
    
    /**
     * @brief Draws the display on the screen
     * @param paused whether to draw the menu or the in game screen
     */
    void draw(bool paused) {
        int ypos = MARGIN*1.2; // the current y position to write at
                               // this is to be incremented
        
        if(paused) { /* Menu screen */
            // Help menu from file
            for(auto str : help_lines) {
                if(str.size() == 0) {
                    ypos += small;
                }
                else if(str[0] == '#') {
                    textColor(sdlRenderer, xpos, ypos, medium, str.substr(1), font_colour);
                    ypos += medium;
                }
                else {
                    textColor(sdlRenderer, xpos, ypos, small, str, font_colour);
                    ypos += small;
                }
            }
            
            // High scores
            ypos += small;
            std::vector<int> const& scores = sc.get_scores();
            textColor(sdlRenderer, xpos, ypos, medium, "High scores", font_colour);
            ypos += medium;
            for(int s : scores) {
                if(s == sc.get_last_score()) {
                    textColor(sdlRenderer, xpos, ypos, small, std::to_string(s) + "     <- last score", 0xc0ffeeFF); 
                }
                else {
                    textColor(sdlRenderer, xpos, ypos, small, std::to_string(s), font_colour);                
                }
                ypos += small;
            }
        }
        else { /* Ingame screen */
            // Current score
            textColor(sdlRenderer, xpos, ypos, large, std::string("Score: ") + std::to_string(sc.get_current_score()), font_colour);
            ypos += large;
            
            // Remaining time
            int time = (int)b.get_time_remaining();
            int minutes = time/60;
            int seconds = time%60;
            std::stringstream ss;
            ss << "Time remaining: "; 
            if(minutes > 0) {
                ss << minutes << "m ";
            }
            ss << seconds << "s";
            
            textColor(sdlRenderer, xpos, ypos, large, ss.str(), (time<10)?0xFF3333FF:font_colour);
            ypos += large/2;
            
            // Event log
            ypos += large;
            std::list<std::pair<int, std::string>>& logs = sc.get_logs();
            for(std::pair<int,std::string>& entry : logs) {
                textColor(sdlRenderer, xpos, ypos, small, entry.second, font_colour);
                ypos += small;
                entry.first++;
            }
            auto i = logs.begin();
            while(i != logs.end() && i->first > 300) {
                i = logs.erase(i);
            }
        }
    }
};

/**
 * @brief Standard SDL2 timer function
 * @param ms the delay between two generated events
 * @param param SDL2 requirement
 * @return the ms parameter (as required)
 */
Uint32 timer(Uint32 ms, void* param) {
    SDL_Event ev;
    ev.type = SDL_USEREVENT;
    SDL_PushEvent(&ev);
    return ms;
}

int main(int argc, char *argv[]) {
    // Intialization of random seed
    srand(time(0));
    
    SDL_Window *sdlWindow;
    try {
        sdlWindow = SDL_CreateWindow("Match",
                                     SDL_WINDOWPOS_CENTERED,
                                     SDL_WINDOWPOS_CENTERED,
                                     WIDTH, HEIGHT,
                                     0);
        if(sdlWindow == NULL) throw 1;
        
        sdlRenderer = SDL_CreateRenderer(sdlWindow, -1, 0);
        if(sdlRenderer == NULL) throw 1;
    }
    catch(int error) {
        SDL_Quit();
        return error;
    }
    
    /* Instantiation of all objects */
    Background bg;
    StatCollector sc;
    Board b(bg, sc);
    Display d(b, sc);
    Selector sel(b);
    
    /* Beginning of time */
    sdlTimer = SDL_AddTimer(20, timer, nullptr);
    music.volume_set(28);
    music.play();
    
    /* Flags for the main event loop */
    bool quit = false;
    bool draw_flag = true;
    bool paused = true;
    
    /* Main event loop */
    while(SDL_WaitEvent(&ev) && !quit) {
        switch(ev.type) {
            case SDL_KEYDOWN:
                switch(ev.key.keysym.sym) {
                    case SDLK_BACKSPACE:
                    case SDLK_DELETE:
                        if(!paused) {
                            sel.unselect();
                        }
                        break;
                    case SDLK_ESCAPE:
                        if(!paused) {
                            if(!sel.unselect()) {
                                sc.commit_score();
                                paused = true;                                
                            }
                        }
                        else {
                            quit = true;
                        }
                        break;
                    case SDLK_m:
                        music.toggle_mute();
                        break;
                    case SDLK_k:
                        sound_success.toggle_mute();
                        break;
                    case SDLK_LEFT:
                    case SDLK_a:
                    case SDLK_RIGHT:
                    case SDLK_d:
                    case SDLK_UP:
                    case SDLK_w:
                    case SDLK_DOWN:
                    case SDLK_s:
                        if(!paused) {
                            sel.move(ev.key.keysym.sym);
                        }
                        break;
                    case SDLK_RETURN:
                    case SDLK_SPACE:
                        if(!paused) {
                            sel.select();
                        }
                        else { /* if paused */
                            paused = false;
                            b.reset();
                        }
                        break;
                }
                break;
            case SDL_USEREVENT:
                draw_flag = !draw_flag;
                if(draw_flag) {
                    bg.draw();
                    d.draw(paused);
                    b.draw();
                    if(!paused) {
                        sel.draw();
                    }
                    SDL_RenderPresent(sdlRenderer);
                }
                
                if(!paused && !b.match_possible()) {
                    sound_magic.play();
                    b.randomly_delete(ROWS*COLS/10);
                }
                
                // Possible fail conditions here
                if(!paused && b.decrease_time()) {
                    sc.commit_score();
                    paused = true;
                }
                break;
                
            case SDL_QUIT:
                quit = true;
                break;
        }
    }
    
    SDL_RemoveTimer(sdlTimer);
    SDL_Quit();
    return 0;
}
